@@ -8,7 +8,7 @@
 
 #include <asm/types.h>          /* For __uXX types */
 
-#define IP_VS_VERSION_CODE            0x000904
+#define IP_VS_VERSION_CODE            0x000905
 #define NVERSION(version)                       \
 	(version >> 16) & 0xFF,                 \
 	(version >> 8) & 0xFF,                  \
@@ -184,8 +184,8 @@ struct ip_vs_dest_user {
 	u_int16_t	port;
 	unsigned	flags;		/* destination flags */
 	int		weight;         /* destination weight */
-	u_int16_t	activeconns;	/* active connections */
-	u_int16_t	inactconns;	/* inactive connections */
+	u_int32_t	activeconns;	/* active connections */
+	u_int32_t	inactconns;	/* inactive connections */
 
 	/* statistics */
 	struct ip_vs_stats_user stats;
@@ -785,6 +785,21 @@ extern __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
 }
 
 
+/*
+ *	TCP/UDP header check (against IP payload size)
+ */
+extern inline int ip_vs_header_check(unsigned proto, char *th, unsigned size)
+{
+	if (proto == IPPROTO_TCP) {
+		unsigned short hlen = ((struct tcphdr*)th)->doff << 2;
+		if (hlen < sizeof(struct tcphdr) || hlen > size)
+			return -1;
+	} else if (proto == IPPROTO_UDP && size < sizeof(struct udphdr))
+		return -1;
+	return 0;
+}
+
+
 /*
  *      Destination cache
  */
