@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.20 2001/09/18 15:42:53 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.21 2001/10/19 15:05:17 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -660,32 +660,34 @@ static int ip_vs_null_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
  */
 static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
-	struct rtable *rt;			/* Route to the other host */
+	struct rtable *rt;		/* Route to the other host */
 	struct iphdr  *iph = skb->nh.iph;
 	union ip_vs_tphdr h;
 	unsigned short size;
-	int doff;                       /* transport protocol data offset */
-	int csum = 0;                   /* checksum */
-	int csum_ok = 0;                /* csum_ok says if csum is valid */
 	int mtu;
 
 	EnterFunction(10);
 
-	h.raw = (char*) iph + iph->ihl * 4;
-	if ((skb = vs_skb_cow(&skb, &iph, &h.raw)) == NULL)
+	/* do more header checking (size) */
+	h.raw = (char*) iph + (iph->ihl << 2);
+	size = ntohs(iph->tot_len) - (iph->ihl << 2);
+	if (ip_vs_header_check(iph->protocol, h.raw, size) == -1)
 		return NF_DROP;
-	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
 
 	/*
-	 *  Must do TCP/UDP checksum checking before mangling the packet
+	 * If it has ip_vs_app helper, the helper may change the payload,
+	 * so it needs full checksum checking and checksum caculation.
+	 * If not, we think the payload checksum is OK, only the header
+	 * (such as IP address and port number) is changed, so it is fast
+	 * to do incremental checksum update, and let the destination host
+	 * do final checksum checking.
 	 */
+
+	/* do TCP/UDP checksum checking if it has application helper */
 	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
 		switch (skb->ip_summed) {
 		case CHECKSUM_NONE:
-			csum = csum_partial(h.raw + doff, size - doff, 0);
-			csum_ok++;
-			skb->csum = csum_partial(h.raw, doff, csum);
+			skb->csum = csum_partial(h.raw, size, 0);
 
 		case CHECKSUM_HW:
 			if (csum_tcpudp_magic(iph->saddr, iph->daddr,
@@ -732,17 +734,9 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	dst_release(skb->dst);
 	skb->dst = &rt->u.dst;
 
-	/*
-	 *	Invalidate csum saving if it has ip_vs_app helper
-	 */
-	csum_ok = !cp->app;
-	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		/* Only port and addr are changed, make fast csum update */
-		ip_vs_fast_check_update(&h, iph->daddr, cp->daddr,
-			h.portp[1], cp->dport, iph->protocol);
-		if (skb->ip_summed == CHECKSUM_HW)
-			skb->ip_summed = CHECKSUM_NONE;
-	}
+	/* copy-on-write the packet before mangling it */
+	if ((skb = vs_skb_cow(&skb, &iph, &h.raw)) == NULL)
+		return NF_DROP;
 
 	/* mangle the packet */
 	iph->daddr = cp->daddr;
@@ -762,34 +756,34 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	}
 
 	/*
-	 * Yug! adjust UDP/TCP checksums
+	 *	Adjust TCP/UDP checksums
 	 */
-	if (!csum_ok) {
-		/* Transport's payload partial checksum */
-		csum = csum_partial(h.raw + doff, size - doff, 0);
+	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
+		/* Only port and addr are changed, make fast csum update */
+		ip_vs_fast_check_update(&h, cp->vaddr, cp->daddr,
+					cp->vport, cp->dport, iph->protocol);
+		if (skb->ip_summed == CHECKSUM_HW)
+			skb->ip_summed = CHECKSUM_NONE;
+	} else {
+		/* full checksum caculation */
+		switch (iph->protocol) {
+		case IPPROTO_TCP:
+			h.th->check = 0;
+			h.th->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
+							size, iph->protocol,
+							csum_partial(h.raw, size, 0));
+			break;
+		case IPPROTO_UDP:
+			h.uh->check = 0;
+			h.uh->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
+							size, iph->protocol,
+							csum_partial(h.raw, size, 0));
+			if (h.uh->check == 0)
+				h.uh->check = 0xFFFF;
+			break;
+		}
 		skb->ip_summed = CHECKSUM_UNNECESSARY;
 	}
-
-	/*
-	 *	Protocol csum
-	 */
-	if (!csum_ok)
-	switch (iph->protocol) {
-	case IPPROTO_TCP:
-		h.th->check = 0;
-		h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-					      size, iph->protocol,
-					      csum_partial(h.raw, doff, csum));
-		break;
-	case IPPROTO_UDP:
-		h.uh->check = 0;
-		h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-					      size, iph->protocol,
-					      csum_partial(h.raw, doff, csum));
-		if (h.uh->check == 0)
-			h.uh->check = 0xFFFF;
-		break;
-	}
 	ip_send_check(iph);
 
 	IP_VS_DBG(10, "NAT to %u.%u.%u.%u:%d\n",
@@ -808,9 +802,9 @@ static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 	LeaveFunction(10);
 	return NF_STOLEN;
 
-  tx_error_icmp:
+ tx_error_icmp:
 	dst_link_failure(skb);
-  tx_error:
+ tx_error:
 	kfree_skb(skb);
 	return NF_STOLEN;
 }
