@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.22 2001/09/18 15:42:54 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.23 2001/10/19 15:05:17 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -64,49 +64,6 @@ EXPORT_SYMBOL(check_for_ip_vs_out);
 /* ID used in ICMP lookups */
 #define icmp_id(icmph)          ((icmph->un).echo.id)
 
-
-/*
- *	Get transport protocol data offset, check against size
- *	return:
- *		0  if other IP proto
- *		-1 if error
- */
-int ip_vs_proto_doff(unsigned proto, char *th, unsigned size)
-{
-	int ret = -1;
-	switch (proto) {
-	case IPPROTO_ICMP:
-		if (size >= sizeof(struct icmphdr))
-			ret = sizeof(struct icmphdr);
-		break;
-	case IPPROTO_UDP:
-		if (size >= sizeof(struct udphdr))
-			ret = sizeof(struct udphdr);
-		break;
-	case IPPROTO_TCP:
-		/*
-		 *	Is this case, this check _also_ avoids
-		 *	touching an invalid pointer if
-		 *	size is invalid
-		 */
-		if (size >= sizeof(struct tcphdr)) {
-			ret = ((struct tcphdr*)th)->doff << 2;
-			if (ret > size) {
-				ret = -1 ;
-			}
-		}
-		break;
-	default:
-		/* Other proto: nothing to say, by now :) */
-		ret = 0;
-	}
-	if (ret < 0)
-		IP_VS_DBG(0, "mess proto_doff for proto=%d, size =%d\n",
-			  proto, size);
-	return ret;
-}
-
-
 const char *ip_vs_proto_name(unsigned proto)
 {
 	static char buf[20];
@@ -547,12 +504,17 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 {
 	struct sk_buff	*skb   = *skb_p;
 	struct iphdr	*iph   = skb->nh.iph;
-	struct icmphdr	*icmph = (struct icmphdr *)((char *)iph + (iph->ihl<<2));
+	struct icmphdr	*icmph;
 	struct iphdr	*ciph;	/* The ip header contained within the ICMP */
 	__u16		*pptr;	/* port numbers from TCP/UDP contained header */
-	struct ip_vs_conn *cp;
-	unsigned short	 len   = ntohs(iph->tot_len) - (iph->ihl * 4);
+	unsigned short	 len;
 	unsigned short	clen, csize;
+	struct ip_vs_conn *cp;
+
+	icmph = (struct icmphdr *)((char *)iph + (iph->ihl<<2));
+	len   = ntohs(iph->tot_len) - (iph->ihl << 2);
+	if (len < sizeof(struct icmphdr))
+		return NF_DROP;
 
 	IP_VS_DBG(12, "outgoing ICMP (%d,%d) %u.%u.%u.%u->%u.%u.%u.%u\n",
 		  icmph->type, ntohs(icmp_id(icmph)),
@@ -662,45 +624,37 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 			      int (*okfn)(struct sk_buff *))
 {
 	struct sk_buff  *skb = *skb_p;
-	struct iphdr	*iph;
+	struct iphdr	*iph = skb->nh.iph;
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
 	int size;
-	int doff;                       /* transport protocol data offset */
-	int csum;                       /* checksum */
-	int csum_ok;                    /* csum_ok says if csum is valid */
 
 	EnterFunction(11);
 
 	if (skb->nfcache & NFC_IPVS_PROPERTY)
 		return NF_ACCEPT;
 
-	iph = skb->nh.iph;
-	h.raw = (char*) iph + iph->ihl * 4;
-	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
+	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
+		skb = ip_defrag(skb);
+		if (!skb) {
+			return NF_STOLEN;
+		}
 
-	switch (doff) {
-	case 0:
-		/*
-		 *	Input path: other IP protos Ok, will
-		 *	reach local sockets path.
-		 */
-		return NF_ACCEPT;
-	case -1:
-		IP_VS_DBG(0, "I-pkt invalid packet data size\n");
-		return NF_DROP;
+		iph = skb->nh.iph;
+		*skb_p = skb;
 	}
 
 	if (iph->protocol == IPPROTO_ICMP)
 		return ip_vs_out_icmp(skb_p);
 
-	IP_VS_DBG(11, "O-pkt: %s size=%d\n",
-		  ip_vs_proto_name(iph->protocol), size);
+	/* let it go if other IP protocols */
+	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
+		return NF_ACCEPT;
 
 	/*
-	 *	Now hunt the list to see if we have an old entry
+	 *	Check if the packet belongs to an old entry
 	 */
+	h.raw = (char*) iph + (iph->ihl << 2);
 	cp = ip_vs_conn_out_get(iph->protocol, iph->saddr, h.portp[0],
 				iph->daddr, h.portp[1]);
 	if (!cp) {
@@ -725,27 +679,28 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		return NF_ACCEPT;
 	}
 
-	/* Check if the checksum is OK, when it is not UDP or check!=0 */
-	csum = 0;
-	csum_ok = 0;
+	/* do more header checking (size) */
+	size = ntohs(iph->tot_len) - (iph->ihl << 2);
+	if (ip_vs_header_check(iph->protocol, h.raw, size) == -1) {
+		ip_vs_conn_put(cp);
+		return NF_DROP;
+	}
+	IP_VS_DBG(11, "O-pkt: %s size=%d\n",
+		  ip_vs_proto_name(iph->protocol), size);
+
+	/*
+	 * If it has ip_vs_app helper, the helper may change the payload,
+	 * so it needs full checksum checking and checksum caculation.
+	 * If not, only the header (addr/port) is changed, so it is fast
+	 * to do incremental checksum update, and let the destination host
+	 * do final checksum checking.
+	 */
+
+	/* do TCP/UDP checksum checking if it has application helper */
 	if (cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		if (skb_is_nonlinear(skb)) {
-			if (skb_linearize(skb, GFP_ATOMIC) != 0) {
-				ip_vs_conn_put(cp);
-				return NF_DROP;
-			}
-			iph = skb->nh.iph;
-			h.raw = (char*) iph + iph->ihl * 4;
-			size = ntohs(iph->tot_len) - (iph->ihl * 4);
-			doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
-		}
 		switch (skb->ip_summed) {
 		case CHECKSUM_NONE:
-			csum = csum_partial(h.raw + doff, size - doff, 0);
-			csum_ok++;
-			IP_VS_DBG(11, "O-pkt: %s I-datacsum=%d\n",
-				  ip_vs_proto_name(iph->protocol), csum);
-			skb->csum = csum_partial(h.raw, doff, csum);
+			skb->csum = csum_partial(h.raw, size, 0);
 		case CHECKSUM_HW:
 			if (csum_tcpudp_magic(iph->saddr, iph->daddr, size,
 					      iph->protocol, skb->csum)) {
@@ -767,23 +722,6 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
-	/*
-	 *	Change the fragments origin
-	 */
-	size = skb->len - (h.raw - skb->nh.raw);
-
-	/*
-	 *	Invalidate csum saving if tunnel has connection helper
-	 */
-	csum_ok = !cp->app;
-	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
-		/* Only port and addr are changed, make fast csum update */
-		ip_vs_fast_check_update(&h, iph->saddr, cp->vaddr,
-			h.portp[0], cp->vport, iph->protocol);
-		if (skb->ip_summed == CHECKSUM_HW)
-			skb->ip_summed = CHECKSUM_NONE;
-	}
-
 	/*
 	 *	Set iph addr and port from ip_vs obj.
 	 */
@@ -802,46 +740,43 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 		iph = skb->nh.iph;
 		h.raw = (char*) iph + iph->ihl *4;
 		size = skb->len - (h.raw - skb->nh.raw);
-		/* doff should have not changed */
 	}
 
 	/*
-	 *	Adjust packet accordingly to protocol
+	 *	Adjust TCP/UDP checksums
 	 */
-	if (!csum_ok) {
-		/* transport's payload partial csum */
-		csum = csum_partial(h.raw + doff, size - doff, 0);
-		skb->ip_summed = CHECKSUM_UNNECESSARY;
-	}
-
-	IP_VS_DBG(11, "O-pkt: %s size=%d O-datacsum=%d\n",
-		  ip_vs_proto_name(iph->protocol), size, csum);
-
-	/*
-	 *	Protocol checksum
-	 */
-	if (!csum_ok)
-	switch (iph->protocol) {
-	case IPPROTO_TCP:
-		h.th->check = 0;
-		h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-					      size, iph->protocol,
-					      csum_partial(h.raw, doff, csum));
-		IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
-			  ip_vs_proto_name(iph->protocol), h.th->check,
-			  (char*) & (h.th->check) - (char*) h.raw);
-		break;
-	case IPPROTO_UDP:
-		h.uh->check = 0;
-		h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-					      size, iph->protocol,
-					      csum_partial(h.raw, doff, csum));
-		if (h.uh->check == 0)
-			h.uh->check = 0xFFFF;
-		IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
-			  ip_vs_proto_name(iph->protocol), h.uh->check,
-			  (char*) &(h.uh->check)- (char*) h.raw);
-		break;
+	if (!cp->app && (iph->protocol != IPPROTO_UDP || h.uh->check != 0)) {
+		/* Only port and addr are changed, make fast csum update */
+		ip_vs_fast_check_update(&h, cp->daddr, cp->vaddr,
+					cp->dport, cp->vport, iph->protocol);
+		if (skb->ip_summed == CHECKSUM_HW)
+			skb->ip_summed = CHECKSUM_NONE;
+	} else {
+		/* full checksum caculation */
+		switch (iph->protocol) {
+		case IPPROTO_TCP:
+			h.th->check = 0;
+			skb->csum = csum_partial(h.raw, size, 0);
+			h.th->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
+							size, iph->protocol,
+							skb->csum);
+			IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
+				  ip_vs_proto_name(iph->protocol), h.th->check,
+				  (char*)&(h.th->check) - (char*)h.raw);
+			break;
+		case IPPROTO_UDP:
+			h.uh->check = 0;
+			skb->csum = csum_partial(h.raw, size, 0);
+			h.uh->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
+							size, iph->protocol,
+							skb->csum);
+			if (h.uh->check == 0)
+				h.uh->check = 0xFFFF;
+			IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
+				  ip_vs_proto_name(iph->protocol), h.uh->check,
+				  (char*)&(h.uh->check) - (char*)h.raw);
+			break;
+		}
 	}
 	ip_send_check(iph);
 
@@ -907,7 +842,9 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 
 	iph = skb->nh.iph;
 	icmph = (struct icmphdr *)((char *)iph+(iph->ihl<<2));
-	len = ntohs(iph->tot_len) - (iph->ihl * 4);
+	len = ntohs(iph->tot_len) - (iph->ihl<<2);
+	if (len < sizeof(struct icmphdr))
+		return NF_DROP;
 
 	IP_VS_DBG(12, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
 		  icmph->type, ntohs(icmp_id(icmph)),
@@ -1059,8 +996,6 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	struct iphdr	*iph = skb->nh.iph;
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
-	unsigned short size;
-	int doff;
 	struct ip_vs_service *svc;
 
 	/*
@@ -1075,36 +1010,21 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		return NF_ACCEPT;
 	}
 
-	h.raw = (char*) iph + iph->ihl * 4;
-
-	/*
-	 *	IP payload size
-	 */
-	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
-
-	switch (doff) {
-	case 0:
-		/*
-		 *	Input path: other IP protos Ok, will
-		 *	reach local sockets path.
-		 */
-		return NF_ACCEPT;
-	case -1:
-		IP_VS_DBG(0, "I-pkt invalid packet data size\n");
-		return NF_DROP;
-	}
-
 	if (iph->protocol == IPPROTO_ICMP)
 		return ip_vs_in_icmp(skb_p);
 
+	/* let it go if other IP protocols */
+	if (iph->protocol != IPPROTO_TCP && iph->protocol != IPPROTO_UDP)
+		return NF_ACCEPT;
+
+	h.raw = (char*) iph + iph->ihl * 4;
 	IP_VS_DBG(11, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
 		  ip_vs_proto_name(iph->protocol),
 		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
 	/*
-	 * reroute to original host:port if found...
+	 * Check if the packet belongs to an existing connection entry
 	 */
 	cp = ip_vs_conn_in_get(iph->protocol, iph->saddr, h.portp[0],
 			       iph->daddr, h.portp[1]);
@@ -1145,7 +1065,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 
 		/*
 		 * Let the virtual server select a real server for the
-		 * incomming connection, and create a connection entry.
+		 * incoming connection, and create a connection entry.
 		 */
 		cp = ip_vs_schedule(svc, iph);
 		if (!cp)
@@ -1326,3 +1246,4 @@ static void __exit ip_vs_cleanup(void)
 
 module_init(ip_vs_init);
 module_exit(ip_vs_cleanup);
+MODULE_LICENSE("GPL");
