@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_sync.c,v 1.2 2001/08/27 02:40:55 wensong Exp $
+ * Version:     $Id: ip_vs_sync.c,v 1.3 2001/09/18 15:42:54 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -284,7 +284,7 @@ static void ip_vs_process_message(const char *buffer, const size_t buflen)
 int ip_vs_sync_state;
 
 /* multicast interface name */
-static char mcast_ifn[IP_VS_IFNAME_MAXLEN];
+char mcast_ifn[IP_VS_IFNAME_MAXLEN];
 
 /* multicast addr */
 static struct sockaddr_in mcast_addr;
