@@ -2,7 +2,7 @@
  *      ipvsadm - IP Virtual Server ADMinistration program
  *                for IPVS NetFilter Module in kernel 2.4
  *
- *      Version: $Id: ipvsadm.c,v 1.14 2001/09/18 15:42:54 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.15 2001/10/19 15:05:17 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *               Peter Kese <peter.kese@ijs.si>
@@ -296,6 +296,7 @@ int main(int argc, char **argv)
 	/* list the table if there is no other arguement */
 	if (argc == 1){
 		list_all(FMT_NONE);
+		ipvs_close();
 		return 0;
 	}
 
@@ -434,7 +435,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		case 'f':
 			set_option(options, OPT_SERVICE);
 			/*
-			 * Set prtocol to a sane values, even
+			 * Set protocol to a sane values, even
 			 * though it is not used
 			 */
 			ur->protocol = IPPROTO_TCP;
@@ -695,7 +696,7 @@ int parse_options(int argc, char **argv, struct ip_vs_rule_user *ur,
 		case 'f':
 			set_option(options, OPT_SERVICE);
 			/*
-			 * Set prtocol to a sane values, even
+			 * Set protocol to a sane values, even
 			 * though it is not used
 			 */
 			ur->protocol = IPPROTO_TCP;
@@ -1422,12 +1423,13 @@ print_service(struct ip_vs_service_user *svc, unsigned int format)
 			sprintf(svc_name, "%s %s",
 				svc->protocol==IPPROTO_TCP?"-t":"-u",
 				vname);
-		else
+		else {
 			sprintf(svc_name, "%s  %s",
 				svc->protocol==IPPROTO_TCP?"TCP":"UDP",
 				vname);
+			svc_name[33] = '\0';
+		}
 		free(vname);
-		svc_name[33] = '\0';
 	}
 
 	/* print virtual service info */
@@ -1478,7 +1480,9 @@ print_service(struct ip_vs_service_user *svc, unsigned int format)
 		if (!(dname = addrport_to_anyname(&daddr, ntohs(e->port),
 						  svc->protocol, format)))
 			exit(1);
-		dname[28] = '\0';
+		if ((!format & FMT_RULE))
+			dname[28] = '\0';
+
 		if (format & FMT_RULE) {
 			printf("-a %s -r %s %s -w %d\n", svc_name, dname,
 			       fwd_switch(e->flags), e->weight);
